#include "inputBase.h"
/*
parameter format for XVDig1394::set_params is a string of "XnXn..."
where X is the parameter character and n is the corresponding value
valid characters and their values:
B : number of buffers ( same as defined in XVVideo.h )
R : pixel format from camera, 0 = YUV422, 1 = RGB, 2 = MONO8,
3 = MONO16, 4 = YUV411, 5 = YUV444
default = 0 
S : scale, 0 = any, 1 = 640x480, 2 = 320x200, 3 = 160x120,
4 = 800x600, 5 = 1024x768, 6 = 1280x960, 7 = 1600x1200
default = 0
M : scale (obsolete), M0 = S4, M1 = S5, M2 = S6, M3 = S7
C : whether to grab center of the image or not, 0 = no, 1 = yes, default = 0
note that C1 implies format 7
T : external trigger, 0 = off, 1-4 = mode 0-3, or directly supply the
number to the camera. default = 0
f : directly set IEEE 1394 camera format (as oppose to set R and S/M)
if the value is 7, S values are still used.
m : directly set IEEE 1394 camera mode   (as oppose to set R and S/M)
note that this parameter should not to set for format 7
r : frame rate, 0 = 1.875, 1 = 3.75, 2 = 7.5, 3 = 15, 4 = 30, 5 = 60 (fps)
default = fastest under selected format and mode
g : gain
u : u component for the white balance
v : v component for the white balance
s : saturation
A : sharpness
h : shutter
a : gamma
x : exposure
o : optical filter
i : bus init (resets the IEEE bus)
*/
/*
parameter format for XVV4L2::set_params is a string of "XnXn..."
where X is the parameter character and n is the corresponding value
valid characters and their values:

I: input
*/
class InputXVSource{
public:
	typedef XV_RGB24 PIX_TYPE24;
	typedef XVImageRGB< PIX_TYPE24 > IMAGE_TYPE24;
	typedef XV_RGB PIX_TYPE32;
	typedef XVImageRGB< PIX_TYPE32 > IMAGE_TYPE32;

	typedef XVVideo< IMAGE_TYPE24 > VID24;
	typedef XVVideo< IMAGE_TYPE32 > VID32;

	typedef XVV4L2< IMAGE_TYPE24 > V4L2;
	typedef XVDig1394< IMAGE_TYPE24 > DIG1394;
	typedef XVImageSeq< IMAGE_TYPE24 > IMG;
	typedef XVMpeg< IMAGE_TYPE32 >  MPG;

	int img_height, img_width;
	virtual void initFrame(int)=0;
	virtual void updateFrame(int)=0;
	virtual void initBuffer(PIX_TYPE24**, IMAGE_TYPE24**)=0;
	virtual void updateBuffer(PIX_TYPE24**)=0;
};

class InputXV24: public InputXVSource{
public:
	VID24 *vid = NULL;
	IMAGE_TYPE24 *vid_buffers;
	int n_buffers;

	InputXV24(char img_source,
		char *dev_path, char *dev_fmt, char* file_path,
		int n_buffers_in, int n_frames) : n_buffers(n_buffers_in){
			switch(img_source) {
			case SRC_IMG: { 			
				fprintf(stdout, "Opening jpeg file at %s with %d frames\n", file_path, n_frames);
				vid = new IMG(file_path, 1, n_frames, n_buffers);
				break;
						  }
			case SRC_USB_CAM: {
				fprintf(stdout, "Opening USB camera %s with format %s\n", dev_path, dev_fmt);
				vid = new V4L2(dev_path, dev_fmt);
				break;
							  }
			case SRC_DIG_CAM: {
				fprintf(stdout, "Opening FIREWIRE camera %s  with format %s\n", dev_path, dev_fmt);
				vid = new DIG1394(dev_path, dev_fmt, DIG1394_NTH_CAMERA(0));
				break;
							  }
			default: {
				fprintf(stdout, "Invalid image source provided for 24 bit Xvision pipeline\n");
				exit(0);
					 }
			}
			XVSize img_size = vid->get_size();
			img_height = img_size.Height();
			img_width = img_size.Width();
	}
	inline void initFrame(int init_buffer_id){
		if(vid->initiate_acquire(init_buffer_id)<0){
			cout<<"Error in InputXV24: Frame could not be acquired\n";
			return;
		}
	}

	inline void initBuffer(PIX_TYPE24 **init_addrs, IMAGE_TYPE24 **buffer_addrs){
		vid_buffers = vid->remap(init_addrs, n_buffers);
		for(int i=0; i<n_buffers;i++){
			vid_buffers[i].resize(img_width, img_height);
			vid_buffers[i].remap((PIX_TYPE24*)buffer_addrs[i]->data(), false);
		}		
	}

	inline void updateFrame(int buffer_id){		
		if(vid->initiate_acquire((buffer_id + 1) % n_buffers)<0){
			cout<<"Error in InputXV24: Frame could not be acquired from the input stream\n";
			return;
		}		
		vid->wait_for_completion(buffer_id);		
	}

	inline void updateBuffer(PIX_TYPE24 **new_addrs){
		for(int i=0; i<n_buffers;i++){
			vid_buffers[i].remap(new_addrs[i], false);
		}
	}
};

class InputXV32: public InputXVSource{

public:
	VID32 *vid = NULL;
	IMAGE_TYPE24 **buffer24;

	int buffer_id32;

	InputXV32(int img_source, char* file_path){

		switch(img_source) {
		case SRC_VID: { 
			fprintf(stdout, "Opening video file %s\n", file_path);
			vid = new MPG(file_path);
			break;
					  }
		default: {
			fprintf(stdout, "Invalid image source provided for 32 bit Xvision pipeline\n");
			exit(0);
				 }
		}
		XVSize img_size=vid->get_size();
		img_height=img_size.Height();
		img_width=img_size.Width();

	}

	inline void initFrame(int init_buffer_id){
		buffer_id32 = init_buffer_id;
		if(vid->initiate_acquire(init_buffer_id)<0){
			cout<<"Error in InputXV32:: Frame could not be acquired\n";
			return;
		}
	}

	inline void initBuffer(PIX_TYPE24 **init_addrs, IMAGE_TYPE24 **buffer_addrs){
		buffer24 = buffer_addrs;
	}

	// XVMpeg does not support 24 bit images so it has to be used to 
	// capture 32 bit images and then copy data to the 24 bit buffers
	inline void updateFrame(int buffer_id24){
		// XVMpeg does not provide any option to set the number of buffers so we have to use 
		// the default hard coded there in XVMpeg.h as 'DEF_MPEG_NUM'
		buffer_id32 = (buffer_id32 + 1) % DEF_MPEG_NUM;
		//printf("**************************getFrameMPG**********************\n");
		if(vid->initiate_acquire(buffer_id32)<0){
			cout<<"Error in InputXV32: Frame could not be acquired from the input stream\n";
			return;
		}
		vid->wait_for_completion(buffer_id32);
		copyFrame32ToBuffer24(vid->frame(buffer_id32), buffer_id24);
	}

	inline void updateBuffer(PIX_TYPE24 **new_addrs){}

	// copy image data from 32 bit source image to 24 bit buffer image
	inline void copyFrame32ToBuffer24(IMAGE_TYPE32 &src_frame, int buffer_id24) {
		uchar* src_data = (uchar*)(src_frame.data());
		uchar* dst_data = (uchar*)(buffer24[buffer_id24]->data());

		for( int row = 0; row < img_height; row++) {
			for ( int col = 0; col < img_width; col++) {
				memcpy(dst_data, src_data, sizeof(PIX_TYPE24));
				dst_data+=sizeof(PIX_TYPE24);
				src_data+=sizeof(PIX_TYPE32);
			}
		}
	}
};

class InputXV:  public InputBase {

public:
	InputXVSource *src;
	InputXV(char img_source=SRC_VID,
		char *dev_name_in=NULL, char *dev_fmt_in=NULL, char* dev_path_in=NULL,
		int n_buffers=1):
	InputBase(img_source, dev_name_in, dev_fmt_in, dev_path_in, n_buffers){
		fprintf(stdout, "Starting InputXV constructor\n");
		if(img_source == SRC_VID){
			src = new InputXV32(img_source, file_path);
		}else{
			src = new InputXV24(img_source, dev_path, dev_fmt, file_path, n_buffers, n_frames);
		}

		img_height = src->img_height;
		img_width = src->img_width;

		fprintf(stdout, "InputXV :: img_height=%d\n", img_height);	
		fprintf(stdout, "InputXV :: img_width=%d\n", img_width);

		PIX_TYPE* data_addrs[n_buffers];
		cv_buffer = new cv::Mat*[n_buffers];
		xv_buffer = new IMAGE_TYPE*[n_buffers];
		for(int i=0; i<n_buffers; i++){
			cv_buffer[i] = new cv::Mat(img_height, img_width, CV_8UC3);
			xv_buffer[i] = new IMAGE_TYPE(img_width, img_height);
			data_addrs[i] = (PIX_TYPE*)(xv_buffer[i]->data());
			cv_buffer[i]->data = (uchar*)xv_buffer[i]->data();
		}
		src->initBuffer(data_addrs, xv_buffer);
		buffer_id = -1;
		updateFrame();
		init_success = 1;
	}

	~InputXV(){
		for(int i=0; i<n_buffers;i++){
			delete(cv_buffer[i]);
			delete(xv_buffer[i]);
		}
		delete(cv_buffer);
		delete(xv_buffer);
	}

	void updateFrame(){
		buffer_id = (buffer_id + 1) % n_buffers;
		src->updateFrame(buffer_id);
		xv_frame = xv_buffer[buffer_id];
		cv_frame = cv_buffer[buffer_id];
		frames_captured++;
	}

	inline void remapBuffer(uchar** new_addr){
		//vid->own_buffers=0;
		//vid_buffers=vid->remap((PIX_TYPE**)new_addr, n_buffers);
		src->updateBuffer((PIX_TYPE**)new_addr);
		for(int i=0;i<n_buffers;i++){
			//vid_buffers[i].resize(img_width, img_height);			
			/*IMAGE_TYPE* temp_frame=&(vid->frame(i));

			//temp_frame->pixmap->own_flag=false;
			temp_frame->resize(img_width, img_height);
			temp_frame->remap((PIX_TYPE*)new_addr[i], false);*/
			xv_buffer[i]->remap((PIX_TYPE*)(new_addr[i], false));
			cv_buffer[i]->data = new_addr[i];
			//temp_frame->pixmap=new XVPixmap<PIX_TYPE>(img_width, img_height, (PIX_TYPE*)new_addr[i], false);
			//temp_frame->win_addr=temp_frame->pixmap->buffer_addr;
		}
		//vid->remap((PIX_TYPE**)new_addr, n_buffers);
		buffer_id = -1;
		src->initFrame(0);
		updateFrame();
	}
};

