#ifndef PROJECTIVE_BASE_H
#define PROJECTIVE_BASE_H

#include "StateSpaceModel.h"

#include <boost/random/linear_congruential.hpp>
#include <boost/random/normal_distribution.hpp>

_MTF_BEGIN_NAMESPACE
// base class for all SSMs that can be expressed 
// by a 3x3 projective transformation matrix
class ProjectiveBase : public StateSpaceModel{

private:
	stringstream err_msg;

	PtsT norm_pts;
	HomPtsT norm_pts_hm;
	CornersT norm_corners;
	HomCornersT norm_corners_hm;

protected:

	ProjWarpT warp_mat, warp_update_mat;
	ProjWarpT inv_warp_mat;

	// let N = no. of pixels and S = no. of state space parameters
	ProjWarpT curr_warp; //! 3 x 3 projective warp matrix 

	// both homogeneous versions are stored too for convenience
	HomPtsT init_pts_hm, curr_pts_hm;
	HomCornersT init_corners_hm, curr_corners_hm;

	const PtsT& getNormPts() const{ return norm_pts; }
	const HomPtsT& getHomNormPts() const{ return norm_pts_hm; }
	const CornersT& getNormCorners() const{ return norm_corners; }
	const HomCornersT& getHomNormCorners() const{ return norm_corners_hm; }

	// get uniformly spaced grid points corresponding to the given corners
	void getPtsFromCorners(ProjWarpT &warp, PtsT &pts, HomPtsT &pts_hm,
		const CornersT &corners);

	// gradient of the SSM w.r.t. its parameters computed at the given point
	// can be used for estimating the change in SSM parameters needed to produce 
	// a required change in point coordinates in image space
	virtual void getInitPixGrad(PtsT &jacobian_prod, int pix_id) = 0;
	virtual void getCurrPixGrad(PtsT &jacobian_prod, int pix_id) = 0;
	// convert the state vector to the 3x3 warp matrix and vice versa
	virtual void getWarpFromState(Matrix3d &warp_mat, const VectorXd& ssm_state) = 0;
	virtual void getStateFromWarp(VectorXd &state_vec, const Matrix3d& warp_mat) = 0;

public:
	// for the overloaded version that takes OpenCV Mat inut
	using StateSpaceModel::setCorners;

	ProjectiveBase(int _resx, int _resy);
	virtual ~ProjectiveBase(){}

	void setCorners(const CornersT& corners) override;
	void setState(const VectorXd &ssm_state) override;
	void additiveUpdate(const VectorXd& state_update) override;
	void invertState(VectorXd& inv_state, const VectorXd& state) override;
	void updateGradPts(double grad_eps) override;
	void updateHessPts(double hess_eps) override;
	void applyWarpToCorners(Matrix24d &warped_corners, const Matrix24d &orig_corners,
		const VectorXd &state_update) override;
	void applyWarpToPts(Matrix2Xd &warped_pts, const Matrix2Xd &orig_pts,
		const VectorXd &state_update) override;
	void applyWarpToPt(double &warped_x, double &warped_y, double x, double y,
		const ProjWarpT &warp);

	// -------------------------------------------------------------------------- //
	// --------------------------- Stochastic Sampler --------------------------- //
	// -------------------------------------------------------------------------- //

	typedef boost::minstd_rand SampleGenT;
	typedef boost::normal_distribution<double> SampleDistT;
	typedef SampleDistT::param_type DistParamT;

	std::vector<SampleGenT> rand_gen;
	std::vector<SampleDistT> rand_dist;
	VectorXd rand_perturbation;

	void initializeSampler(const VectorXd &state_sigma) override;
	void initializeSampler(double pix_sigma) override;
	void setSamplerMean(const VectorXd &mean) override;
	void setSamplerSigma(const VectorXd &std) override;
	void setSamplerSigma(double pix_sigma) override;

	VectorXd getSamplerSigma() override;
	VectorXd getSamplerMean() override;

	void additiveRandomWalk(VectorXd &perturbed_state,
		const VectorXd &base_state) override;
	void compositionalRandomWalk(VectorXd &perturbed_state,
		const VectorXd &base_state) override;

	void additiveAutoRegression1(VectorXd &perturbed_state, VectorXd &perturbed_ar,
		const VectorXd &base_state, const VectorXd &base_ar, double a = 0.5) override;
	void compositionalAutoRegression1(VectorXd &perturbed_state, VectorXd &perturbed_ar,
		const VectorXd &base_state, const VectorXd &base_ar, double a = 0.5) override;

	void generatePerturbation(VectorXd &perturbation) override;
	void generatePerturbedPts(VectorXd &perturbed_pts) override;
	void estimateMeanOfSamples(VectorXd &sample_mean,
		const std::vector<VectorXd> &samples, int n_samples) override;
	void getPerturbedPts(VectorXd &perturbed_pts,
		const VectorXd &state_perturbation) override;
	void estimateStateSigma(VectorXd &state_sigma, double pix_sigma) override;
};
_MTF_END_NAMESPACE

#endif
