#ifndef MTF_NT_H
#define MTF_NT_H

//! External interface of the MTF library; this is the only header that needs to be included to use the framework
//! Provides functions to create trackers corresponding to different combinations of 
//! search methods, appearance models and state space models as well as some third party 

#include "opencv2/core/core.hpp"

// parameters for the different modules
#include "mtf/Config/parameters.h"

// Non templated implementations of search methods
#include "mtf/SM/NT/FCLK.h"

// appearance models
#include "mtf/AM/SSD.h"
#include "mtf/AM/NSSD.h"
#include "mtf/AM/ZNCC.h"
#include "mtf/AM/SCV.h"
#include "mtf/AM/LSCV.h"
#include "mtf/AM/RSCV.h"
#include "mtf/AM/LRSCV.h"
#include "mtf/AM/KLD.h"
#include "mtf/AM/LKLD.h"	
#include "mtf/AM/MI.h"
#include "mtf/AM/SPSS.h"
#include "mtf/AM/SSIM.h"
#include "mtf/AM/NCC.h"
#include "mtf/AM/CCRE.h"
#include "mtf/AM/FMaps.h"

// state space models
#include "mtf/SSM/LieHomography.h"
#include "mtf/SSM/CornerHomography.h"
#include "mtf/SSM/HomographySL3.h"
#include "mtf/SSM/Homography.h"
#include "mtf/SSM/Affine.h"
#include "mtf/SSM/Similitude.h"
#include "mtf/SSM/Isometry.h"
#include "mtf/SSM/Transcaling.h"
#include "mtf/SSM/Translation.h"

_MTF_BEGIN_NAMESPACE

using namespace params;

inline ImageBase *getPixMapperObj(const char *pix_mapper_type, ImgParams *img_params){
	if(!pix_mapper_type)
		return nullptr;
	if(!strcmp(pix_mapper_type, "ssd")){
		return new SSD(img_params);
	} else if(!strcmp(pix_mapper_type, "nssd")){
		NSSDParams *nssd_params = new NSSDParams(img_params, norm_pix_max, norm_pix_min, debug_mode);
		return new NSSD(nssd_params);
	} else if(!strcmp(pix_mapper_type, "zncc")){
		return new ZNCC(img_params);
	} else if(!strcmp(pix_mapper_type, "scv")){
		SCVParams *scv_params = new SCVParams(img_params, scv_use_bspl, scv_n_bins, scv_preseed, scv_pou,
			scv_weighted_mapping, scv_mapped_gradient, debug_mode);
		return new SCV(scv_params);
	} else if(!strcmp(pix_mapper_type, "lscv")){
		LSCVParams *lscv_params = new LSCVParams(img_params, lscv_sub_regions, lscv_sub_regions,
			lscv_spacing, lscv_spacing, scv_affine_mapping, scv_once_per_frame, scv_n_bins, scv_preseed,
			scv_weighted_mapping,  lscv_show_subregions, debug_mode);
		return new LSCV(lscv_params);
	} else if(!strcmp(pix_mapper_type, "lrscv") || !strcmp(pix_mapper_type, "lrsc")){
		LRSCVParams *lrscv_params = new LRSCVParams(img_params, lscv_sub_regions, lscv_sub_regions,
			lscv_spacing, lscv_spacing, scv_affine_mapping, scv_once_per_frame, scv_n_bins, scv_preseed,
			scv_weighted_mapping,  lscv_show_subregions, debug_mode);
		return new LRSCV(lrscv_params);
	} else if(!strcmp(pix_mapper_type, "rscv")){
		RSCVParams *rscv_params = new RSCVParams(img_params, scv_use_bspl, scv_n_bins, scv_preseed, scv_pou,
			scv_weighted_mapping, scv_mapped_gradient, debug_mode);
		return new RSCV(rscv_params);
	} else{
		throw std::invalid_argument("getPixMapperObj::Invalid pixel mapper type provided");
	}
}

inline StateSpaceModel *getSSM(const char *ssm_type){
	if(!strcmp(ssm_type, "lie_hom") || !strcmp(ssm_type, "l8")){
		LieHomographyParams *lhomm_params = new LieHomographyParams();
		lhomm_params->normalized_init = hom_normalized_init;
		return new LieHomography(resx, resy, lhomm_params);
	} else if(!strcmp(ssm_type, "chom") || !strcmp(ssm_type, "c8")){
		CornerHomographyParams *chom_params = new CornerHomographyParams();
		chom_params->normalized_init = hom_normalized_init;
		chom_params->grad_eps = grad_eps;
		return new mtf::CornerHomography(resx, resy, chom_params);
	} else if(!strcmp(ssm_type, "lhom2") || !strcmp(ssm_type, "l82")){
		HomographySL3Params *hom_params = new HomographySL3Params(hom_normalized_init, debug_mode);
		return new mtf::HomographySL3(resx, resy, hom_params);
	} else if(!strcmp(ssm_type, "hom") || !strcmp(ssm_type, "8")){
		HomographyParams *hom_params = new HomographyParams(hom_normalized_init, hom_corner_based_sampling, debug_mode);
		return new mtf::Homography(resx, resy, hom_params);
	} else if(!strcmp(ssm_type, "aff") || !strcmp(ssm_type, "6")){
		AffineParams *aff_params = new AffineParams();
		aff_params->normalized_init = hom_normalized_init;
		return new mtf::Affine(resx, resy, aff_params);
	} else if(!strcmp(ssm_type, "sim") || !strcmp(ssm_type, "4")){
		return new mtf::Similitude(resx, resy);
	} else if(!strcmp(ssm_type, "iso") || !strcmp(ssm_type, "3")){
		return new mtf::Isometry(resx, resy);
	} else if(!strcmp(ssm_type, "trs") || !strcmp(ssm_type, "3s")){
		return new mtf::Transcaling(resx, resy);
	} else if(!strcmp(ssm_type, "trans") || !strcmp(ssm_type, "2")){
		return new mtf::Translation(resx, resy);
	} else{
		printf("Invalid state space model provided: %s\n", ssm_type);
		return nullptr;
	}
}
inline AppearanceModel *getAM(const char *am_type){
	ImgParams img_params(resx, resy, grad_eps, hess_eps);
	if(!strcmp(am_type, "ssd")){
		return new SSD(&img_params);
	} else if(!strcmp(am_type, "nssd")){
		NSSDParams *nssd_params = new NSSDParams(&img_params, norm_pix_max, norm_pix_min, debug_mode);
		return new NSSD(nssd_params);
	} else if(!strcmp(am_type, "zncc")){
		return new ZNCC(&img_params);
	} else if(!strcmp(am_type, "scv")){
		SCVParams *scv_params = new SCVParams(&img_params, scv_use_bspl, scv_n_bins, scv_preseed, scv_pou,
			scv_weighted_mapping, scv_mapped_gradient, debug_mode);
		return new SCV(scv_params);
	} else if(!strcmp(am_type, "lscv")){
		LSCVParams *lscv_params = new LSCVParams(&img_params, lscv_sub_regions, lscv_sub_regions,
			lscv_spacing, lscv_spacing, scv_affine_mapping, scv_once_per_frame, scv_n_bins, scv_preseed,
			scv_weighted_mapping, lscv_show_subregions, debug_mode);
		return new LSCV(lscv_params);
	} else if(!strcmp(am_type, "rscv")){
		RSCVParams *rscv_params = new RSCVParams(&img_params, scv_use_bspl, scv_n_bins, scv_preseed, scv_pou,
			scv_weighted_mapping, scv_mapped_gradient, debug_mode);
		return new RSCV(rscv_params);
	} else if(!strcmp(am_type, "lrscv") || !strcmp(am_type, "lrsc")){
		LRSCVParams *lscv_params = new LRSCVParams(&img_params, lscv_sub_regions, lscv_sub_regions,
			lscv_spacing, lscv_spacing, scv_affine_mapping, scv_once_per_frame, scv_n_bins, scv_preseed,
			scv_weighted_mapping,  lscv_show_subregions, debug_mode);
		return new LRSCV(lscv_params);
	} else if(!strcmp(am_type, "kld")){
		KLDParams *kld_params = new KLDParams(&img_params, mi_n_bins, mi_pre_seed, mi_pou, debug_mode);
		return new KLD(kld_params);
	} else if(!strcmp(am_type, "lkld")){
		LKLDParams *lkld_params = new LKLDParams(&img_params, lkld_sub_regions, lkld_sub_regions,
			lkld_spacing, lkld_spacing, lkld_n_bins, lkld_pre_seed, lkld_pou, debug_mode);
		return new LKLD(lkld_params);
	} else if(!strcmp(am_type, "mi")){
		MIParams *mi_params = new MIParams(&img_params, mi_n_bins, mi_pre_seed, mi_pou,
			getPixMapperObj(pix_mapper, &img_params), debug_mode);
		return new MI(mi_params);
	} else if(!strcmp(am_type, "spss")){
		SPSSParams *spss_params = new SPSSParams(&img_params, ssim_k1,
			getPixMapperObj(pix_mapper, &img_params));
		return new SPSS(spss_params);
	} else if(!strcmp(am_type, "ssim")){
		SSIMParams *ssim_params = new SSIMParams(&img_params, ssim_k1, ssim_k2);
		return new SSIM(ssim_params);
	} else if(!strcmp(am_type, "ncc")){
		NCCParams *ncc_params = new NCCParams(&img_params, ncc_fast_hess);
		return new NCC(ncc_params);
	} else if(!strcmp(am_type, "ccre")){
		CCREParams *ccre_params = new CCREParams(&img_params, ccre_n_bins, ccre_pou, ccre_pre_seed,
			ccre_symmetrical_grad, ccre_n_blocks, debug_mode);
		return new CCRE(ccre_params);
	} else if(!strcmp(am_type, "fmaps")){
		return new FMaps(&img_params);
	} else{
		printf("Invalid appearance model provided: %s\n", am_type);
		return nullptr;
	}
}

inline nt::SearchMethod *getSM(const char *sm_type,
	const char *am_type, const char *ssm_type){
	AppearanceModel *am = getAM(am_type);
	StateSpaceModel *ssm = getSSM(ssm_type);
	if(!strcmp(sm_type, "fclk") || !strcmp(sm_type, "fc")){
		nt::FCLKParams *fclk_params = new nt::FCLKParams(max_iters, epsilon,
			static_cast<nt::FCLKParams::HessType>(hess_type), sec_ord_hess,
			update_templ, fc_chained_warp, debug_mode);
		return new nt::FCLK(am, ssm, fclk_params);
	} else{
		printf("Invalid search method provided: %s\n", sm_type);
		return nullptr;
	}
}
_MTF_END_NAMESPACE

#endif